// f.olofsson2023
// 20 OSC controlled PWMs
// Pins 16-19 are bit-banged. They have jitter and a max frequency of around 1kHz.

// Serial protocol:
// [150, 151, 152, nn, nn, nn, nn, nn, nn, nn, nn, nn, nn, nn, nn, nn, nn, nn, nn, nn, nn, nn, nn, 153]
// where nn is 0 - 255
// OSC protocol:
// /pin, index, float (0.0 - 1.0)
// /pins, ff, ff, ff, ff, ff, ff, ff, ff, ff, ff, ff, ff, ff, ff, ff, ff, ff, ff, ff, ff
// where ff is 0.0 - 1.0
// /pwm, index, int (0 - 100)
// /pwms, ii, ii, ii, ii, ii, ii, ii, ii, ii, ii, ii, ii, ii, ii, ii, ii, ii, ii, ii, ii
// where ii is 0 - 100

#include <Arduino.h>
#include <WiFi.h>
#include <WiFiUdp.h>
#include <OSCMessage.h>
#include <OSCData.h>

//--settings

#define ID 0                               // EDIT give unique ID to each ESP you program
const char *ssid = "myssid";               // EDIT your accessPoint network name
const char *password = "mypass";           // EDIT your password
const char *hostname = "f0pwm";            // EDIT hostname
const uint16_t out_port = 17123;           // EDIT default output OSC port
const uint16_t in_port = 17122;            // EDIT default input OSC port
const uint32_t defaultFrequency = 1000;    // EDIT Hz - can also be set via OSC

//--

const u_int8_t pins[] = {
    32, 33, 25, 26, 27, 14, 12, 13, // left hand side
    15, 2, 0, 4, 16, 17, 5, 18,     // right hand side (2 is also the blue LED)
    19, 21, 22, 23                  // last four right hand side, bit-banged
};
const u_int8_t resolution = 10;
const u_int16_t pwmMax = pow(2, resolution) - 1;
u_int32_t frequency = defaultFrequency;
IPAddress out_ip;
WiFiUDP Udp;
u_int16_t pwm[20];
u_int8_t serialData[20];
u_int8_t serialCounter = 0;
u_int32_t bitBangDelay = 0;
TaskHandle_t PWMBitBangTask;

void pwmBitBangTask(void *parameter) {
  u_int16_t pwmCounter = 0;
  unsigned long nextTime = 0;
  while (1) {
    unsigned long now = micros();
    if (now >= nextTime) {
      nextTime = now + bitBangDelay;
      for (u_int8_t i = 16; i < 20; i++) {
        if (pwmCounter == 0) {
          digitalWrite(pins[i], pwm[i] >= pwmMax);
        } else if (pwm[i] == (pwmMax - pwmCounter)) {
          digitalWrite(pins[i], 1);
        }
      }
      pwmCounter = (pwmCounter + 1) % pwmMax;
    }
  }
}

void setup() {

  for (u_int8_t i = 0; i < 20; i++) {
    if (i < 16) {
      ledcSetup(i, frequency, resolution);
      ledcAttachPin(pins[i], i);
      ledcWrite(i, 0);
    } else {
      pinMode(pins[i], OUTPUT);
      digitalWrite(pins[i], 0);
    }
    pwm[i] = 0;
  }

  delay(10);

  Serial.begin(57600);

  xTaskCreate(
    pwmBitBangTask,
    "PWMBitBangTask",
    1000,
    NULL,
    1,
    NULL
  );

  WiFi.setHostname(hostname);
  WiFi.mode(WIFI_STA);
  WiFi.config(INADDR_NONE, INADDR_NONE, INADDR_NONE, INADDR_NONE);
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
  }
  Udp.begin(in_port);
  out_ip = WiFi.localIP();
  out_ip[3] = 255; // broadcast ping
}

void pwmWrite() {
  for (u_int8_t i = 0; i < 16; i++) {
    ledcWrite(i, pwm[i]);
  }
}

void setPin(u_int8_t i, float value) {
  pwm[constrain(i, 0, 19)] = constrain(value, 0.0, 1.0) * pwmMax;
}

void setPwm(u_int8_t i, float value) {
  pwm[constrain(i, 0, 19)] = constrain(value, 0.0, 100.0) * 0.01 * pwmMax;
}

float getFloatCast(OSCMessage &msg, u_int8_t index) {
  if (msg.isInt(index)) {
    return float(msg.getInt(index));
  }
  return msg.getFloat(index);
}

void oscPang(OSCMessage &m) {
  OSCMessage msg("/f0pwm_ping");
  msg.add(ID);
  msg.add(WiFi.localIP()[0]);
  msg.add(WiFi.localIP()[1]);
  msg.add(WiFi.localIP()[2]);
  msg.add(WiFi.localIP()[3]);
  msg.add(20); // number of pwm
  msg.add(frequency);
  msg.add(resolution);
  Udp.beginPacket(out_ip, out_port);
  msg.send(Udp);
  Udp.endPacket();
  msg.empty();
}

void oscPin(OSCMessage &msg) {
  setPin(msg.getInt(0), getFloatCast(msg, 1));
}

void oscPins(OSCMessage &msg) {
  for (u_int8_t i = 0; i < 20; i++) {
    setPin(i, getFloatCast(msg, i));
  }
}

void oscPwm(OSCMessage &msg) {
  u_int8_t i = msg.getInt(0);
  if ((i >= 0) && (i < 20)) {
    setPwm(i, getFloatCast(msg, 1));
  }
}

void oscPwms(OSCMessage &msg) {
  for (u_int8_t i = 0; i < 20; i++) {
    setPwm(i, getFloatCast(msg, i));
  }
}

void oscFreq(OSCMessage &msg) {
  frequency = max(1, msg.getInt(0));
  bitBangDelay = round(1000000.0 / (pwmMax * frequency));
  for (u_int8_t i = 0; i < 16; i++) {
    ledcSetup(i, frequency, resolution);
  }
}

void loop() {

  int packet_size = Udp.parsePacket();
  if (packet_size > 0) {
    OSCMessage msg;
    while (packet_size--) {
      msg.fill(Udp.read());
    }
    if (!msg.hasError()) {
      msg.dispatch("/pin", oscPin);   // 0.0 - 1.0
      msg.dispatch("/pins", oscPins); // array of 0.0 - 1.0
      msg.dispatch("/pwm", oscPwm);   // 0 - 100
      msg.dispatch("/pwms", oscPwms); // array of 0 - 100
      msg.dispatch("/freq", oscFreq); // set frequency
      msg.dispatch("/pang", oscPang); // reply with ping + IP

      pwmWrite();
    }
  }

  if (Serial.available() > 0) {
    u_int8_t b = Serial.read();
    if ((serialCounter == 0) && (b == 150)) {
      serialCounter = 1;
    } else if ((serialCounter == 1) && (b == 151)) {
      serialCounter = 2;
    } else if ((serialCounter == 2) && (b == 152)) {
      serialCounter = 3;
    } else if (serialCounter >= 3 && (serialCounter <= 22)) {
      serialData[serialCounter - 3] = b;
      serialCounter++;
    } else if (serialCounter == 23 && (b == 153)) {
      for (u_int8_t i = 0; i < 20; i++) {
        setPin(i, serialData[i] / 255.0);
      }
      pwmWrite();
      serialCounter = 0;
    } else {
      serialCounter = 0;
    }
  }

  delay(2);
}
